/*
 *  This file contains a simple ICMP ping client.
 *
 *  COPYRIGHT (c) 2005.
 *  LOYTEC electronics GmbH, 2005.
 *
 *  Thomas Rauscher <trauscher@loytec.com>
 *
 *  The license and distribution terms for this file may be
 *  found in the file LICENSE in this distribution or at
 *  http://www.OARcorp.com/rtems/license.html.
 *
 */

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/time.h>

#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "ping.h"

rtems_ping_cfg_t rtems_ping_standard =
{
    {1, 0},           /* Timeout (1s) */
    64-ICMP_MINLEN    /* Payload      */
};

static void rtems_ping_pattern(rtems_ping_t *ping)
{
  unsigned i;
  int n;
  unsigned8 *p = (unsigned8 *)ping->icmp_req->icmp_data;

  /* Write payload pattern */
  for(i=0, n=0x80+ping->seq; i<ping->payload_size; i++, n++)
  {
    *p++ = n;
  }
}

static unsigned16 rtems_ping_cksum(unsigned16 *p, int len)
{
  unsigned32 accu = 0;

  /* IP checksum */
  
  for(; len>1; len-=2)
  {
    accu += *p++;
  }

  if (len == 1) {
    accu += *p & 0xff00;
  }

  accu = (accu >> 16) + (accu & 0xffff);
  accu += (accu >> 16);
  return((~accu) & 0x0000ffff);
}

rtems_ping_t *rtems_ping_open(unsigned32              ip_addr,
			      const rtems_ping_cfg_t *cfg)
{
  rtems_ping_t *ping;
  size_t        payload_size;

  /* Check if to use standard config. */
  if(!cfg)
  {
    cfg = &rtems_ping_standard;
  }

  /* Determine payload size. */
  payload_size = cfg->payload_size;
  if(payload_size < 64 - ICMP_MINLEN)
  {
    payload_size = 64 - ICMP_MINLEN;
  }

  /* Get memory for ping control structure. */
  ping = calloc(sizeof(*ping), 1);
  if(!ping)
  {
    rtems_ping_close(ping);
    return NULL;
  }

  /* Get memory for packets */
  ping->req_size     = ((unsigned char *)ping->icmp_req->icmp_data - 
			(unsigned char *)ping->icmp_req) + payload_size;
  ping->resp_size    = IP_MAXPACKET;
  ping->raw_req      = calloc(ping->req_size, 1);
  ping->raw_resp     = calloc(ping->resp_size, 1);
  ping->payload_size = payload_size;
  ping->timeout      = cfg->timeout;
  ping->seq          = rand();

  if(!ping->raw_req || !ping->raw_resp)
  {
    rtems_ping_close(ping);
    return NULL;
  }
  
  /* Open RAW ICMP socket */
  ping->socket = socket(AF_INET, SOCK_RAW, 1);
  if(ping->socket == -1)
  {
    rtems_ping_close(ping);
    return NULL;
  }

  /* Set destination address */
  memset(&ping->dest, 0, sizeof(ping->dest));
  ping->dest.sin_family = AF_INET;
  ping->dest.sin_addr.s_addr = htonl(ip_addr);

  /* Initialize request */
  ping->icmp_req->icmp_type  = ICMP_ECHO;
  ping->icmp_req->icmp_code  = 0;
  ping->icmp_req->icmp_id    = (int) ping;

  return ping;
}

int rtems_ping_send(rtems_ping_t *ping, rtems_interval *trip_time)
{
  int                retval;
  int                sent;
  int                rcvd;
  int                srclen;
  struct sockaddr_in src;
  rtems_interval     send_time;
  rtems_interval     rcv_time;
  n_short            id = (n_short) (((unsigned long) ping) & 0xffff);
  struct timeval     tv;
  unsigned long      tps = _TOD_Ticks_per_second;
  rtems_interval     timeout;

  if(!ping) {
    return EINVAL;
  }

  /* Increment sequence number */
  ping->seq++;

  /* Set values for this ping ... */
  rtems_ping_pattern(ping);
  ping->icmp_req->icmp_cksum = 0;
  ping->icmp_req->icmp_seq   = ping->seq;
  ping->icmp_req->icmp_id    = id;

  /* ... and update checksum. */
  ping->icmp_req->icmp_cksum = rtems_ping_cksum((unsigned16*)ping->raw_req,
						ping->req_size);

  /* Get time for ping. */
  (void) rtems_clock_get(RTEMS_CLOCK_GET_TICKS_SINCE_BOOT, &send_time);

  rcv_time = send_time; /* For first timeout calculation */
  timeout = send_time + (ping->timeout.tv_sec * tps) +
    ((ping->timeout.tv_usec * tps) / 1000000);

  /* Send ping. */
  sent = sendto(ping->socket, ping->raw_req, ping->req_size, 0,
		(struct sockaddr *)&ping->dest, sizeof(ping->dest));
  
  /* Wait for pong. */
  while(1) 
  {
    if(rcv_time >= timeout) 
    {
      retval = ETIMEDOUT;
      break;
    }

    rcv_time = timeout - rcv_time;
    /* Set timeout */
    tv.tv_sec  = (rcv_time / tps);
    tv.tv_usec = (rcv_time % tps) * _TOD_Microseconds_per_tick;
    if(setsockopt(ping->socket, SOL_SOCKET,
		  SO_RCVTIMEO, (char*)&(tv),  sizeof(tv))<0)
    {
      retval = -1;
      break;
    }

    /* Wait for response. */
    srclen = sizeof(src);
    memset(ping->raw_resp, 0, ping->req_size);
    rcvd = recvfrom(ping->socket, ping->raw_resp, ping->resp_size, 0, 
		    (struct sockaddr *)&src, &srclen);
    
    /* Get ICMP response */
    ping->icmp_resp = ((struct icmp *)
		       ((unsigned char*)ping->raw_resp +
			ping->raw_resp->ip_hl * 4));

    if(rcvd < 0)
    {
      /* Nothing received ... */
      retval = ETIMEDOUT;
      break;
    }

    /* Packet received, remember receive time */
    (void) rtems_clock_get(RTEMS_CLOCK_GET_TICKS_SINCE_BOOT, &rcv_time);

    /* Check if packet is response to our ping */
    if((ping->icmp_resp->icmp_type == ICMP_ECHOREPLY) &&
       (ping->icmp_resp->icmp_id == id) && 
       (ping->icmp_resp->icmp_seq == ping->icmp_req->icmp_seq) &&
       (src.sin_addr.s_addr == ping->dest.sin_addr.s_addr))
    {
      if(trip_time)
      {
	*trip_time = (rtems_interval)((int)rcv_time - (int)send_time);
      }
      /* Check if payload matches. */
      retval = (memcmp(ping->icmp_req->icmp_data,
		       ping->icmp_resp->icmp_data +
		       (((struct ip *) ping->icmp_resp)->ip_hl<<2),
		       ping->payload_size)==0) ? 0 : EIO;
      break;
    }
  }

  /* Done. */
  return retval;
}

int rtems_ping_close(rtems_ping_t *ping)
{
  int retval = 0;
  
  if(ping)
  {
    /* Close socket. */
    if(ping->socket > 2)
    {
      retval = close(ping->socket);
    }
    
    /* Free memory. */
    free(ping->raw_resp);
    free(ping->raw_req);
    free(ping);
  }

  return retval;
}
